@@ -6,11 +6,14 @@ use core::f64;
 
 use hashbrown::HashMap;
 use pyo3::prelude::*;
+use pyo3::types::PyTuple;
 
 use crate::bytecode;
-use crate::error::{message_bad_eof, message_from_token, message_incorrect_requirement};
+use crate::error::{
+    message_bad_eof, message_from_token, message_incorrect_requirement, QASM2ParseError,
+};
 use crate::lex::{Token, TokenContext, TokenStream, TokenType};
-use crate::parse::GateSymbol;
+use crate::parse::{GateSymbol, GlobalSymbol};
 
 /// Enum representation of the builtin OpenQASM 2 functions.  The built-in Qiskit parser adds the
 /// inverse trigonometric functions, but these are an extension to the version as given in the
@@ -107,6 +110,7 @@ enum Atom {
     LParen,
     RParen,
     Function(Function),
+    CustomFunction(PyObject, usize),
     Op(Op),
     Const(f64),
     Parameter(usize),
@@ -127,6 +131,7 @@ pub enum Expr {
     Divide(Box<Expr>, Box<Expr>),
     Power(Box<Expr>, Box<Expr>),
     Function(Function, Box<Expr>),
+    CustomFunction(PyObject, Vec<Expr>),
 }
 
 impl IntoPy<PyObject> for Expr {
@@ -174,6 +179,11 @@ impl IntoPy<PyObject> for Expr {
                 argument: expr.into_py(py),
             }
             .into_py(py),
+            Expr::CustomFunction(func, exprs) => bytecode::ExprCustom {
+                callable: func,
+                arguments: exprs.into_iter().map(|expr| expr.into_py(py)).collect(),
+            }
+            .into_py(py),
         }
     }
 }
@@ -216,6 +226,8 @@ pub struct ExprParser<'a> {
     pub tokens: &'a mut Vec<TokenStream>,
     pub context: &'a mut TokenContext,
     pub gate_symbols: &'a HashMap<String, GateSymbol>,
+    pub global_symbols: &'a HashMap<String, GlobalSymbol>,
+    pub strict: bool,
 }
 
 impl<'a> ExprParser<'a> {
@@ -253,31 +265,41 @@ impl<'a> ExprParser<'a> {
     /// Expect a token of the correct [TokenType].  This is a direct analogue of
     /// [parse::State::expect].  The error variant of the result contains a suitable error message
     /// if the expectation is violated.
-    fn expect(
-        &mut self,
-        expected: TokenType,
-        required: &str,
-        cause: &Token,
-    ) -> Result<Token, String> {
+    fn expect(&mut self, expected: TokenType, required: &str, cause: &Token) -> PyResult<Token> {
         let token = match self.next_token() {
-            None => return Err(message_bad_eof(self.current_filename(), required, cause)),
+            None => {
+                return Err(QASM2ParseError::new_err(message_bad_eof(
+                    self.current_filename(),
+                    required,
+                    cause,
+                )))
+            }
             Some(token) => token,
         };
         if token.ttype == expected {
             Ok(token)
         } else {
-            Err(message_incorrect_requirement(
+            Err(QASM2ParseError::new_err(message_incorrect_requirement(
                 self.current_filename(),
                 required,
                 &token,
-            ))
+            )))
+        }
+    }
+
+    /// Peek the next token from the stream, and consume and return it only if it has the correct
+    /// type.
+    fn accept(&mut self, acceptable: TokenType) -> Option<Token> {
+        match self.peek_token() {
+            Some(Token { ttype, .. }) if *ttype == acceptable => self.next_token(),
+            _ => None,
         }
     }
 
     /// Apply a prefix [Op] to the current [expression][Expr].  If the current expression is a
     /// constant floating-point value the application will be eagerly constant-folded, otherwise
     /// the resulting [Expr] will have a tree structure.
-    fn apply_prefix(&mut self, prefix: Op, expr: Expr) -> Result<Expr, String> {
+    fn apply_prefix(&mut self, prefix: Op, expr: Expr) -> PyResult<Expr> {
         match prefix {
             Op::Plus => Ok(expr),
             Op::Minus => match expr {
@@ -291,20 +313,14 @@ impl<'a> ExprParser<'a> {
     /// Apply a binary infix [Op] to the current [expression][Expr].  If both operands have
     /// constant floating-point values the application will be eagerly constant-folded, otherwise
     /// the resulting [Expr] will have a tree structure.
-    fn apply_infix(
-        &mut self,
-        infix: Op,
-        lhs: Expr,
-        rhs: Expr,
-        op_token: &Token,
-    ) -> Result<Expr, String> {
+    fn apply_infix(&mut self, infix: Op, lhs: Expr, rhs: Expr, op_token: &Token) -> PyResult<Expr> {
         if let (Expr::Constant(val), Op::Divide) = (&rhs, infix) {
             if *val == 0.0 {
-                return Err(message_from_token(
+                return Err(QASM2ParseError::new_err(message_from_token(
                     op_token,
                     "cannot divide by zero",
                     self.current_filename(),
-                ));
+                )));
             }
         };
         if let (Expr::Constant(val_l), Expr::Constant(val_r)) = (&lhs, &rhs) {
@@ -333,12 +349,7 @@ impl<'a> ExprParser<'a> {
     /// Apply a "scientific calculator" built-in function to an [expression][Expr].  If the operand
     /// is a constant, the function will be constant-folded to produce a new constant expression,
     /// otherwise a tree-form [Expr] is returned.
-    fn apply_function(
-        &mut self,
-        func: Function,
-        expr: Expr,
-        token: &Token,
-    ) -> Result<Expr, String> {
+    fn apply_function(&mut self, func: Function, expr: Expr, token: &Token) -> PyResult<Expr> {
         match expr {
             Expr::Constant(val) => match func {
                 Function::Cos => Ok(Expr::Constant(val.cos())),
@@ -347,14 +358,14 @@ impl<'a> ExprParser<'a> {
                     if val > 0.0 {
                         Ok(Expr::Constant(val.ln()))
                     } else {
-                        Err(message_from_token(
+                        Err(QASM2ParseError::new_err(message_from_token(
                             token,
                             &format!(
                                 "failure in constant folding: cannot take ln of non-positive {}",
                                 val
                             ),
                             self.current_filename(),
-                        ))
+                        )))
                     }
                 }
                 Function::Sin => Ok(Expr::Constant(val.sin())),
@@ -362,14 +373,14 @@ impl<'a> ExprParser<'a> {
                     if val >= 0.0 {
                         Ok(Expr::Constant(val.sqrt()))
                     } else {
-                        Err(message_from_token(
+                        Err(QASM2ParseError::new_err(message_from_token(
                             token,
                             &format!(
                                 "failure in constant folding: cannot take sqrt of negative {}",
                                 val
                             ),
                             self.current_filename(),
-                        ))
+                        )))
                     }
                 }
                 Function::Tan => Ok(Expr::Constant(val.tan())),
@@ -378,11 +389,73 @@ impl<'a> ExprParser<'a> {
         }
     }
 
+    fn apply_custom_function(
+        &mut self,
+        callable: PyObject,
+        exprs: Vec<Expr>,
+        token: &Token,
+    ) -> PyResult<Expr> {
+        if exprs.iter().all(|x| matches!(x, Expr::Constant(_))) {
+            // We can still do constant folding with custom user classical functions, we're just
+            // going to have to acquire the GIL and call the Python object the user gave us right
+            // now.  We need to explicitly handle any exceptions that might occur from that.
+            Python::with_gil(|py| {
+                let args = PyTuple::new(
+                    py,
+                    exprs.iter().map(|x| {
+                        if let Expr::Constant(val) = x {
+                            *val
+                        } else {
+                            unreachable!()
+                        }
+                    }),
+                );
+                match callable.call1(py, args) {
+                    Ok(retval) => match retval.extract::<f64>(py) {
+                        Ok(fval) => Ok(Expr::Constant(fval)),
+                        Err(inner) => {
+                            let error = QASM2ParseError::new_err(message_from_token(
+                                token,
+                                "user-defined function returned non-float during constant folding",
+                                self.current_filename(),
+                            ));
+                            error.set_cause(py, Some(inner));
+                            Err(error)
+                        }
+                    },
+                    Err(inner) => {
+                        let error = QASM2ParseError::new_err(message_from_token(
+                            token,
+                            "caught exception when constant folding with user-defined function",
+                            self.current_filename(),
+                        ));
+                        error.set_cause(py, Some(inner));
+                        Err(error)
+                    }
+                }
+            })
+        } else {
+            Ok(Expr::CustomFunction(callable, exprs))
+        }
+    }
+
+    /// If in `strict` mode, and we have a trailing comma, emit a suitable error message.
+    fn check_trailing_comma(&self, comma: Option<&Token>) -> PyResult<()> {
+        match (self.strict, comma) {
+            (true, Some(token)) => Err(QASM2ParseError::new_err(message_from_token(
+                token,
+                "[strict] trailing commas in parameter and qubit lists are forbidden",
+                self.current_filename(),
+            ))),
+            _ => Ok(()),
+        }
+    }
+
     /// Convert the given general [Token] into the expression-parser-specific [Atom], if possible.
     /// Not all [Token]s have a corresponding [Atom]; if this is the case, the return value is
     /// `Ok(None)`.  The error variant is returned if the next token is grammatically valid, but
     /// not semantically, such as an identifier for a value of an incorrect type.
-    fn try_atom_from_token(&self, token: &Token) -> Result<Option<Atom>, String> {
+    fn try_atom_from_token(&self, token: &Token) -> PyResult<Option<Atom>> {
         match token.ttype {
             TokenType::LParen => Ok(Some(Atom::LParen)),
             TokenType::RParen => Ok(Some(Atom::RParen)),
@@ -404,18 +477,31 @@ impl<'a> ExprParser<'a> {
                 let id = token.text(self.context);
                 match self.gate_symbols.get(id) {
                     Some(GateSymbol::Parameter { index }) => Ok(Some(Atom::Parameter(*index))),
-                    Some(GateSymbol::Qubit { .. }) => Err(message_from_token(
-                        token,
-                        &format!("'{}' is a gate qubit, not a parameter", id),
-                        self.current_filename(),
-                    )),
-                    None => Err(message_from_token(
-                        token,
-                        &format!("'{}' is not a parameter defined in this scope", id),
-                        self.current_filename(),
-                    )),
+                    Some(GateSymbol::Qubit { .. }) => {
+                        Err(QASM2ParseError::new_err(message_from_token(
+                            token,
+                            &format!("'{}' is a gate qubit, not a parameter", id),
+                            self.current_filename(),
+                        )))
+                    }
+                    None => {
+                        match self.global_symbols.get(id) {
+                            Some(GlobalSymbol::Classical { callable, n_params }) => {
+                                Ok(Some(Atom::CustomFunction(callable.clone(), *n_params)))
+                            }
+                            _ =>  {
+                            Err(QASM2ParseError::new_err(message_from_token(
+                            token,
+                            &format!(
+                                "'{}' is not a parameter or custom instruction defined in this scope",
+                                id,
+                            ),
+                            self.current_filename(),)))
+                        }
+                    }
                 }
             }
+            }
             _ => Ok(None),
         }
     }
@@ -442,9 +528,9 @@ impl<'a> ExprParser<'a> {
     /// expression), then as many `*` and `^` operations as appear would be evaluated by this loop,
     /// and its parsing would finish when it saw the next `+` binary operation.  For initial entry,
     /// the `power_min` should be zero.
-    fn eval_expression(&mut self, power_min: u8, cause: &Token) -> Result<Expr, String> {
+    fn eval_expression(&mut self, power_min: u8, cause: &Token) -> PyResult<Expr> {
         let token = self.next_token().ok_or_else(|| {
-            message_bad_eof(
+            QASM2ParseError::new_err(message_bad_eof(
                 self.current_filename(),
                 if power_min == 0 {
                     "an expression"
@@ -452,10 +538,10 @@ impl<'a> ExprParser<'a> {
                     "a missing operand"
                 },
                 cause,
-            )
+            ))
         })?;
         let atom = self.try_atom_from_token(&token)?.ok_or_else(|| {
-            message_incorrect_requirement(
+            QASM2ParseError::new_err(message_incorrect_requirement(
                 self.current_filename(),
                 if power_min == 0 {
                     "an expression"
@@ -463,7 +549,7 @@ impl<'a> ExprParser<'a> {
                     "a missing operand"
                 },
                 &token,
-            )
+            ))
         })?;
         // First evaluate the "left-hand side" of a (potential) sequence of binary infix operators.
         // This might be a simple value, a unary operator acting on a value, or a bracketed
@@ -479,36 +565,73 @@ impl<'a> ExprParser<'a> {
             }
             Atom::RParen => {
                 if power_min == 0 {
-                    Err(message_from_token(
+                    Err(QASM2ParseError::new_err(message_from_token(
                         &token,
                         "did not find an expected expression",
                         self.current_filename(),
-                    ))
+                    )))
                 } else {
-                    Err(message_from_token(
+                    Err(QASM2ParseError::new_err(message_from_token(
                         &token,
                         "the parenthesis closed, but there was a missing operand",
                         self.current_filename(),
-                    ))
+                    )))
                 }
             }
             Atom::Function(func) => {
                 let lparen_token =
                     self.expect(TokenType::LParen, "an opening parenthesis", &token)?;
                 let argument = self.eval_expression(0, &token)?;
+                let comma = self.accept(TokenType::Comma);
+                self.check_trailing_comma(comma.as_ref())?;
                 self.expect(TokenType::RParen, "a closing parenthesis", &lparen_token)?;
                 Ok(self.apply_function(func, argument, &token)?)
             }
+            Atom::CustomFunction(callable, n_params) => {
+                let lparen_token =
+                    self.expect(TokenType::LParen, "an opening parenthesis", &token)?;
+                let mut arguments = Vec::<Expr>::with_capacity(n_params);
+                let mut comma = None;
+                loop {
+                    // There are breaks at the start and end of this loop, because we might be
+                    // breaking because there are _no_ parameters, because there's a trailing
+                    // comma before the closing parenthesis, or because we didn't see a comma after
+                    // an expression so we _need_ a closing parenthesis.
+                    if let Some((Atom::RParen, _)) = self.peek_atom() {
+                        break;
+                    }
+                    arguments.push(self.eval_expression(0, &token)?);
+                    comma = self.accept(TokenType::Comma);
+                    if comma.is_none() {
+                        break;
+                    }
+                }
+                self.check_trailing_comma(comma.as_ref())?;
+                self.expect(TokenType::RParen, "a closing parenthesis", &lparen_token)?;
+                if arguments.len() == n_params {
+                    Ok(self.apply_custom_function(callable, arguments, &token)?)
+                } else {
+                    Err(QASM2ParseError::new_err(message_from_token(
+                        &token,
+                        &format!(
+                            "custom function argument-count mismatch: expected {}, saw {}",
+                            n_params,
+                            arguments.len(),
+                        ),
+                        self.current_filename(),
+                    )))
+                }
+            }
             Atom::Op(op) => match prefix_power(op) {
                 Some(power) => {
                     let expr = self.eval_expression(power, &token)?;
                     Ok(self.apply_prefix(op, expr)?)
                 }
-                None => Err(message_from_token(
+                None => Err(QASM2ParseError::new_err(message_from_token(
                     &token,
                     &format!("'{}' is not a valid unary operator", op.text()),
                     self.current_filename(),
-                )),
+                ))),
             },
             Atom::Const(val) => Ok(Expr::Constant(val)),
             Atom::Parameter(val) => Ok(Expr::Parameter(val)),
@@ -531,7 +654,7 @@ impl<'a> ExprParser<'a> {
 
     /// Parse a single expression completely. This is the only public entry point to the
     /// operator-precedence parser.
-    pub fn parse_expression(&mut self, cause: &Token) -> Result<Expr, String> {
+    pub fn parse_expression(&mut self, cause: &Token) -> PyResult<Expr> {
         self.eval_expression(0, cause)
     }
 }
